/*
 *				**** GetInfo Loc ****
 *
 * This is the implementation for GetInfo Location package
 *
 * The contructor returns a handle rather than a pointer
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "getinfo.h"

/*********************************************
 *
 * Private class variables
 *           
 * By declaring the following variables as static,
 * they are available to any routine within this source file
 * however, they cannot be accessed by any routines outside
 * of this file.
 *
 ********************************************/

static int debugsw = 0;		/* debug switch */
/*
 * this is used to designate the maximum number of entries
 * We can allow the user to modify this with a special routine
 */
static int max_ent = INFO_MAX_ENT;
/******************************************************
 *
 * This is the actual list
 * We could use a static sized array for simplicity
 * (but why would I do that? :^)
 * Instead, I am including the ability to have the user
 * specify the maximum size of the collection
 *
 * In real life, I would probably use a ProdPtrList
 * from my ProdList package.  That way, I could
 * allow the user to increase the max size as desired
 *
 * Regardless of how you implement the list, you may want
 * to enforce a maximum number of entries
 * otherwise, the user can crash your package by calling
 * the create_info() routine inside an infinite loop
 ********************************************************/
static Info **List = NULL; /* an array of pointers */

/**************** what the static array would look like:
static Info *List[INFO_MAX_ENT]; /* list of entries */

/**************** what the ProdList implementation would look like:
static ProdPtrList *List = NULL; /* list of entities */

/**********************************************
*
* Private class methods
*
* By declaring these routines as static,
* only other routines within this source file
* can call them.
*
* This gives us the same effect as a private class method
* in a true object oriented language
*
************************************************/

void info_set_max(int n)
{ /* must be called first or it won't have any affect */
	if (List == NULL && n >= 0) max_ent = n;	/* new max number of entries */
}

void free_info_ent(Info *ent)
{ /* free entry */
	if (ent != NULL) free(ent);
}

Info *create_info_ent(void)
{ /*create entry */
	Info *ent;

   ent = (Info *)malloc(sizeof(Info));
   if (ent != NULL)
   { /* initialize entry */
   	ent->acct = INFO_BAD_ACCT;
      ent->fname = NULL;
      ent->lname = NULL;
      ent->bal = INFO_BAD_BAL;
   }
   return ent;
}

void info_terminate(void)
{ /* clean up when done */
	int i;
   Info *ent;

   if (List == NULL) return;	/* nothing to clean up */
   for (i=0; i < max_ent; i++)
   { /* release all entries */
		ent = List[i];	/* get entry */
      if (ent != NULL) free_info_ent(ent);
   }
   free(List); 	/* free list */
   List = NULL;	/* clear pointer */
}

static void init(void)
{ /* called first time - initialize class */
	int i;

	/* allocate the list */
   List = (Info **)malloc(sizeof(Info) * max_ent);

   /* clear all entries in the array */
   for (i=0; i < max_ent; i++) List[i] = NULL;
}

static int get_free_info(void)
{ /* find a free entry */
	int i;
   Info *ent;

   if (List == NULL) init();	/* check for first time */

   /* find unused entry (the pointer value will be NULL) */
   for (i=0; i < max_ent; i++)
   {
   	if (List[i] == NULL)
      { /* free entry found - fill it */
			if ((ent = create_info_ent()) == NULL) return INFO_BAD_ENT;
		   /* we have an entry, store it in the list */
		   List[i] = ent;
         return i;	/* location of new entry */
      }
   }
	return INFO_BAD_ENT; /* list is full */
}

static Info *get_info(int h)
{ /* return entry at location 'h' */
   if (List == NULL) init();	/* check for first time */

	if (h < 0 || h > max_ent) return NULL;	/* bad args */
   return List[h];	/* return pointer to entry */
}

/***********************************************************
 *
 * Public Routines
 *
 * These routines are available by any other source
 * file within the project.
 *
 * This gives us the same functionality as a public
 * class method in a true object oriented language.
 *
 * Each routine validates the input parameters before
 * processing the request
 *		If a handle is passed, we make sure it is a valid handle
 *		If a field value is passed, we make sure it is valid
 *
 ************************************************************/

int create_info(void)
{
	int ret;

   ret = get_free_info();	/* find free location */
   return ret;		/* return that location */
}

void free_info(int h)
{
	Info *ent;

   ent = get_info(h);
	if (ent != NULL)
   {
   	free_info_ent(ent);
      List[h] = NULL;
   }
}

/* get/set routines for fields */
int get_info_acct(int h)
{
	Info *ent;		/* info entry */

   ent = get_info(h);
   if (ent == NULL) return INFO_BAD_ACCT;	/* bad return */
   return ent->acct;	/* return account number */
}

void set_info_acct(int h, int acct)
{
	Info *ent;

   ent = get_info(h);
   /* account number should be positive */
   if (ent != NULL && acct > 0) ent->acct = acct;
}


/*****************************************************************
 *
 * Get first name for entry
 *
 * There are three ways we can do this:
 *	(1) we can return the actual string
 *		If we do this, the user can change the value directly
 *		because we have given the user the actual address
 * (2) we can return a copy of the string that the user must free
 *		If we do this, and the user forgets to free the returned string,
 *		we will have memory leaks
 * (3) we can return a copy of the string from a static area
 *		If we do this, each new call to this routine will clobber
 *		the same area.  This means that if the user wants to store
 *		this value for some time, they should make their own copy
 *
 *****************************************************************/
char *get_info_fname(int h)
{
	Info *ent;
   static char ret[BUFSIZ];	/* return string */

   ent = get_info(h);
   if (ent == NULL) return NULL;

   /*
    * valid request - return name
    * each option is shown below.
    * All but they one we are using is commented out
    */

   /********************
    option 1:
   	return ent->fname;
   ********************/

   /*********************
     option 2:
     	return strdup(ent->fname);
   ************************/

   /**********************
   	option 3:
   ***********************/
   	return strcpy(ret, ent->fname); /* possible buffer overflow */
}

/*************************************************************
 * set first name of entry
 *
 * To prevent the user from being able to alter the field
 * after this routine returns, we make a copy of the string
 * that we are passed
 *
 **************************************************************/
void set_info_fname(int h, char *fname)
{
	Info *ent;

   ent = get_info(h);
   /* make sure handle was valid. Field has to be non-NULL and not too big */
   if (ent != NULL && fname != NULL && strlen(fname) < BUFSIZ)
   { /* valid request - change fname */
   	if (ent->fname != NULL) free(ent->fname);	/* free old string */
      ent->fname = strdup(fname);	/* make copy of name */
   }
}

char *get_info_lname(int h)
{
	Info *ent;
   static char ret[BUFSIZ];	/* return string */

   ent = get_info(h);
   if (ent == NULL) return NULL;

   return strcpy(ret, ent->lname);
}

void set_info_lname(int h, char *lname)
{
	Info *ent;

   ent = get_info(h);
   /* make sure handle was valid. Field has to be non-NULL and not too big */
   if (ent != NULL && lname != NULL && strlen(lname) < BUFSIZ)
   { /* valid request - change fname */
   	if (ent->lname != NULL) free(ent->lname);	/* free old string */
      ent->lname = strdup(lname);	/* make copy of name */
   }
}

float get_info_bal(int h)
{
	Info *ent;

   ent = get_info(h);
   if (ent != NULL) return ent->bal;
   else return INFO_BAD_BAL;	/* bad balance */
}

void set_info_bal(int h, float bal)
{
	Info *ent;

   ent = get_info(h);
   if (ent != NULL && bal >= 0.0) ent->bal = bal;
}

/*
 * utility routine to display a (possibly empty) string
 */
static char *
get_info_string(char *s)
{
	if (s == NULL) return "[empty]";
   return s;
}

void dump_info(int h, FILE *fp)
{
	Info *ent;
   char *s;

   ent = get_info(h);
   if (ent == NULL || fp == NULL) return;	/* invalid args */
   fprintf(fp, "\nEntry %d:\nAcct:\t%d\n", h, ent->acct);
   s = get_info_string(ent->fname);
	fprintf(fp, "Fname:\t>>%s<<\n", s);
	s = get_info_string(ent->lname);
   fprintf(fp, "Lname:\t>>%s<<\n", s);
   fprintf(fp, "Bal:\t%.2f\n", ent->bal);
}

void display_info(int h)
{ /* display entry  content */
	dump_info(h, stdout);
}

